"""
TODO:
 - auto_proxy_init can capture `import as` by scanning globals for sentinels
  - replaces _AutoProxyInitCapture._attrs dict outright
  - should raise on unbound or shadowed import - was probably imported for side-effects but will never get
    proxy imported
"""
import builtins
import contextlib
import functools
import importlib.util
import types
import typing as ta

from ..lazyglobals import LazyGlobals


##


class NamePackage(ta.NamedTuple):
    name: str
    package: str


class _ProxyInit:
    class _Import(ta.NamedTuple):
        pkg: str
        attr: str | None

    def __init__(
            self,
            lazy_globals: LazyGlobals,
            name_package: NamePackage,
    ) -> None:
        super().__init__()

        self._lazy_globals = lazy_globals
        self._name_package = name_package

        self._imps_by_attr: dict[str, _ProxyInit._Import] = {}
        self._mods_by_pkgs: dict[str, ta.Any] = {}

    @property
    def name_package(self) -> NamePackage:
        return self._name_package

    def add(
            self,
            package: str,
            attrs: ta.Iterable[str | tuple[str, str]] | None = None,
    ) -> None:
        if isinstance(attrs, str):
            raise TypeError(attrs)

        if attrs is None:
            whole_attr = package.split('.')[-1]

            self._imps_by_attr[whole_attr] = self._Import(package, None)
            self._lazy_globals.set_fn(whole_attr, functools.partial(self.get, whole_attr))

        else:
            for attr in attrs:
                if isinstance(attr, tuple):
                    imp_attr, attr = attr
                else:
                    imp_attr = attr

                self._imps_by_attr[attr] = self._Import(package, imp_attr)
                self._lazy_globals.set_fn(attr, functools.partial(self.get, attr))

    def _import_module(self, name: str) -> ta.Any:
        return importlib.import_module(name, package=self._name_package.package)

    def get(self, attr: str) -> ta.Any:
        try:
            imp = self._imps_by_attr[attr]
        except KeyError:
            raise AttributeError(attr)  # noqa

        val: ta.Any

        if imp.attr is None:
            val = self._import_module(imp.pkg)

        else:
            try:
                mod = self._mods_by_pkgs[imp.pkg]
            except KeyError:
                mod = self._import_module(imp.pkg)
                self._mods_by_pkgs[imp.pkg] = mod

            val = getattr(mod, imp.attr)

        return val


def proxy_init(
        init_globals: ta.MutableMapping[str, ta.Any],
        package: str,
        attrs: ta.Iterable[str | tuple[str, str]] | None = None,
) -> None:
    if isinstance(attrs, str):
        raise TypeError(attrs)

    init_name_package = NamePackage(
        init_globals['__name__'],
        init_globals['__package__'],
    )

    pi: _ProxyInit
    try:
        pi = init_globals['__proxy_init__']

    except KeyError:
        pi = _ProxyInit(
            LazyGlobals.install(init_globals),
            init_name_package,
        )
        init_globals['__proxy_init__'] = pi

    else:
        if pi.name_package != init_name_package:
            raise Exception(f'Wrong init name: {pi.name_package=} != {init_name_package=}')

    pi.add(package, attrs)


##


class AutoProxyInitError(Exception):
    pass


class AutoProxyInitErrors:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    class HookError(AutoProxyInitError):
        pass

    class AttrError(AutoProxyInitError):
        def __init__(self, module: str | None, name: str) -> None:
            super().__init__()

            self.module = module
            self.name = name

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(module={self.module!r}, name={self.name!r})'

    class ImportError(AutoProxyInitError):  # noqa
        def __init__(self, module: str, from_list: ta.Sequence[str] | None) -> None:
            super().__init__()

            self.module = module
            self.from_list = from_list

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(module={self.module!r}, from_list={self.from_list!r})'

    class ImportStarForbiddenError(ImportError):
        pass

    class UnproxiedImportForbiddenError(ImportError):
        pass


class _AutoProxyInitCapture:
    class ModuleSpec(ta.NamedTuple):
        name: str
        level: int

        def __str__(self) -> str:
            return f'{"." * self.level}{self.name}'

        def __repr__(self) -> str:
            return repr(str(self))

    class _ModuleAttr:
        def __init__(self, module: '_AutoProxyInitCapture._Module', name: str) -> None:
            super().__init__()

            self.__module = module
            self.__name = name

        def __repr__(self) -> str:
            return f'<{self.__class__.__name__}: {f"{self.__module.spec}:{self.__name}"!r}>'

    class _Module:
        def __init__(self, spec: '_AutoProxyInitCapture.ModuleSpec') -> None:
            super().__init__()

            self.spec = spec

            self.module = types.ModuleType(f'<{self.__class__.__qualname__}: {spec!r}>')

            self.attrs: dict[str, _AutoProxyInitCapture._ModuleAttr] = {}
            self.imported_whole = False

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}({self.spec!r})'

    def __init__(self) -> None:
        super().__init__()

        self._modules: dict[_AutoProxyInitCapture.ModuleSpec, _AutoProxyInitCapture._Module] = {}
        self._attrs: dict[str, _AutoProxyInitCapture._ModuleAttr | _AutoProxyInitCapture._Module] = {}

    def _handle_import(
            self,
            module: _Module,
            *,
            from_list: ta.Sequence[str] | None,
    ) -> None:
        if from_list is None:
            if module.spec.level or not module.spec.name:
                raise AutoProxyInitError

            attr = module.spec.name

            try:
                xma: ta.Any = self._attrs[attr]
            except KeyError:
                pass

            else:
                if (
                        xma is not self._attrs.get(attr) or
                        not module.imported_whole
                ):
                    raise AutoProxyInitErrors.AttrError(str(module.spec), attr)

                return

            self._attrs[attr] = module
            module.imported_whole = True

        else:
            for attr in from_list:
                if attr == '*':
                    raise AutoProxyInitErrors.ImportStarForbiddenError(str(module.spec), from_list)

                try:
                    xma = getattr(module.module, attr)
                except AttributeError:
                    pass

                else:
                    if (
                            xma is not module.attrs.get(attr) or
                            xma is not self._attrs.get(attr)
                    ):
                        raise AutoProxyInitErrors.AttrError(str(module.spec), attr)

                    continue

                if attr in self._attrs:
                    raise AutoProxyInitErrors.AttrError(str(module.spec), attr)

                ma = _AutoProxyInitCapture._ModuleAttr(module, attr)
                self._attrs[attr] = ma
                module.attrs[attr] = ma
                setattr(module.module, attr, ma)

    _MOD_SELF_ATTR: ta.ClassVar[str] = '__auto_proxy_init_capture__'

    def _intercept_import(
            self,
            name: str,
            *,
            globals: ta.Mapping[str, ta.Any] | None = None,  # noqa
            from_list: ta.Sequence[str] | None = None,
            level: int = 0,
    ) -> types.ModuleType | None:
        if not (
                globals is not None and
                globals.get(self._MOD_SELF_ATTR) is self
        ):
            return None

        spec = _AutoProxyInitCapture.ModuleSpec(name, level)
        try:
            module = self._modules[spec]
        except KeyError:
            module = self._Module(spec)
            self._modules[spec] = module

        self._handle_import(
            module,
            from_list=from_list,
        )

        return module.module

    @contextlib.contextmanager
    def hook_context(
            self,
            init_globals: ta.MutableMapping[str, ta.Any],  # noqa
            *,
            forbid_unproxied_imports: bool = False,
    ) -> ta.Iterator[None]:
        if self._MOD_SELF_ATTR in init_globals:
            raise AutoProxyInitErrors.HookError

        old_import = builtins.__import__

        def new_import(
                name,
                globals=None,  # noqa
                locals=None,  # noqa
                fromlist=None,
                level=0,
        ):
            if (im := self._intercept_import(
                    name,
                    globals=globals,
                    from_list=fromlist,
                    level=level,
            )) is not None:
                return im

            if forbid_unproxied_imports:
                raise AutoProxyInitErrors.UnproxiedImportForbiddenError(
                    str(_AutoProxyInitCapture.ModuleSpec(name, level)),
                    fromlist,
                )

            return old_import(
                name,
                globals=globals,
                locals=locals,
                fromlist=fromlist,
                level=level,
            )

        #

        init_globals[self._MOD_SELF_ATTR] = self
        builtins.__import__ = new_import

        try:
            yield

        finally:
            if not (
                    init_globals[self._MOD_SELF_ATTR] is self and
                    builtins.__import__ is new_import
            ):
                raise AutoProxyInitErrors.HookError

            del init_globals[self._MOD_SELF_ATTR]
            builtins.__import__ = old_import

    def verify_globals(
            self,
            init_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> None:
        for attr, obj in self._attrs.items():
            try:
                xo = init_globals[attr]
            except KeyError:
                raise AutoProxyInitErrors.AttrError(None, attr) from None

            if isinstance(obj, _AutoProxyInitCapture._ModuleAttr):
                if xo is not obj:
                    raise AutoProxyInitErrors.AttrError(None, attr) from None

            elif isinstance(obj, _AutoProxyInitCapture._Module):
                if xo is not obj.module:
                    raise AutoProxyInitErrors.AttrError(None, attr) from None

            else:
                raise TypeError(obj)

    @property
    def all_attrs(self) -> ta.AbstractSet[str]:
        return self._attrs.keys()

    class ProxyInit(ta.NamedTuple):
        package: str
        attrs: ta.Sequence[str] | None

    def build_proxy_inits(self) -> list[ProxyInit]:
        lst: list[_AutoProxyInitCapture.ProxyInit] = []

        for module in self._modules.values():
            if module.imported_whole:
                lst.append(_AutoProxyInitCapture.ProxyInit(str(module.spec), None))

            if module.attrs:
                if not module.spec.name:
                    for attr in module.attrs:
                        if not module.spec.level:
                            raise AutoProxyInitError

                        lst.append(_AutoProxyInitCapture.ProxyInit('.' * module.spec.level + attr, None))

                else:
                    lst.append(_AutoProxyInitCapture.ProxyInit(str(module.spec), list(module.attrs)))

        return lst


@contextlib.contextmanager
def auto_proxy_init(
        init_globals: ta.MutableMapping[str, ta.Any],
        *,
        disable: bool = False,
        eager: bool = False,
) -> ta.Iterator[None]:
    """
    This is a bit extreme - use sparingly. It relies on an interpreter-global import lock, but much of the ecosystem
    implicitly does anyway. It further relies on temporarily patching `__builtins__.__import__`, but could be switched
    to use any number of other import hooks.
    """

    if disable:
        yield
        return

    cap = _AutoProxyInitCapture()

    with cap.hook_context(init_globals):
        yield

    cap.verify_globals(init_globals)

    pis = cap.build_proxy_inits()

    for attr in cap.all_attrs:
        del init_globals[attr]

    for pi in pis:
        proxy_init(
            init_globals,
            pi.package,
            pi.attrs,
        )

    if eager:
        lg = LazyGlobals.install(init_globals)
        for attr in cap.all_attrs:
            lg.get(attr)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
Should be kept somewhat lightweight - used in cli entrypoints.

TODO:
 - persisted caching support - {pkg_name: manifests}
 - real relative cls names - shouldn't need parent package names
 - *require* loaded class names - special All sentinel for explicit all
  - ! late instantiation !
 - TypeMap style weak cache of issubclass queries
  - wait.. lazily load the class for virtual subclass queries? xor support virtual bases?
"""
import dataclasses as dc
import importlib.machinery
import importlib.resources
import importlib.util
import json
import os.path
import threading
import typing as ta

from .types import Manifest


##


class ManifestLoader:
    class LoadedManifest:
        def __init__(
                self,
                package: 'ManifestLoader.LoadedPackage',
                manifest: Manifest,
        ) -> None:
            super().__init__()

            self._package = package
            self._manifest = manifest

        @property
        def package(self) -> 'ManifestLoader.LoadedPackage':
            return self._package

        @property
        def manifest(self) -> Manifest:
            return self._manifest

        @property
        def loader(self) -> 'ManifestLoader':
            return self._package.loader

        @property
        def class_key(self) -> str:
            [(cls_key, value_dct)] = self._manifest.value.items()
            return cls_key

        _value: ta.Any

        def value(self) -> ta.Any:
            try:
                return self._value
            except AttributeError:
                pass

            value = self.loader._instantiate_loaded_manifest(self)  # noqa
            self._value = value
            return value

    class LoadedPackage:
        def __init__(
                self,
                loader: 'ManifestLoader',
                name: str,
        ) -> None:
            super().__init__()

            self._loader = loader
            self._name = name

        _manifests: ta.Sequence['ManifestLoader.LoadedManifest']

        @property
        def loader(self) -> 'ManifestLoader':
            return self._loader

        @property
        def name(self) -> str:
            return self._name

        @property
        def manifests(self) -> ta.Sequence['ManifestLoader.LoadedManifest']:
            return self._manifests

    def __init__(
            self,
            *,
            module_remap: ta.Optional[ta.Mapping[str, str]] = None,
            value_instantiator: ta.Optional[ta.Callable[..., ta.Any]] = None,
    ) -> None:
        super().__init__()

        self._value_instantiator = value_instantiator
        self._module_remap = module_remap or {}

        self._lock = threading.RLock()

        self._module_reverse_remap = {v: k for k, v in self._module_remap.items()}

        self._loaded_classes: ta.Dict[str, type] = {}
        self._loaded_packages: ta.Dict[str, ta.Optional[ManifestLoader.LoadedPackage]] = {}

        self._scanned_package_root_dirs: ta.Dict[str, ta.Sequence[str]] = {}

    #

    @classmethod
    def kwargs_from_entry_point(
            cls,
            globals: ta.Mapping[str, ta.Any],  # noqa
            *,
            module_remap: ta.Optional[ta.Mapping[str, str]] = None,
            **kwargs: ta.Any,
    ) -> ta.Dict[str, ta.Any]:
        rm: ta.Dict[str, str] = {}

        if module_remap:
            rm.update(module_remap)

        if '__name__' in globals and '__spec__' in globals:
            name: str = globals['__name__']
            spec: importlib.machinery.ModuleSpec = globals['__spec__']
            if '__main__' not in rm and name == '__main__':
                rm[spec.name] = '__main__'

        return dict(module_remap=rm, **kwargs)

    #

    def _load_class_uncached(self, key: str) -> type:
        if not key.startswith('$'):
            raise Exception(f'Bad key: {key}')

        parts = key[1:].split('.')
        pos = next(i for i, p in enumerate(parts) if p[0].isupper())

        mod_name = '.'.join(parts[:pos])
        mod_name = self._module_remap.get(mod_name, mod_name)
        mod = importlib.import_module(mod_name)

        obj: ta.Any = mod
        for ca in parts[pos:]:
            obj = getattr(obj, ca)

        cls = obj
        if not isinstance(cls, type):
            raise TypeError(cls)

        return cls

    def _load_class_locked(self, key: str) -> type:
        try:
            return self._loaded_classes[key]
        except KeyError:
            pass

        cls = self._load_class_uncached(key)
        self._loaded_classes[key] = cls
        return cls

    def _load_class(self, key: str) -> type:
        with self._lock:
            return self._load_class_locked(key)

    #

    def _deserialize_raw_manifests(self, obj: ta.Any, pkg_name: str) -> ta.Sequence[Manifest]:
        if not isinstance(obj, (list, tuple)):
            raise TypeError(obj)

        lst: ta.List[Manifest] = []
        for e in obj:
            m = Manifest(**e)

            m = dc.replace(m, module=pkg_name + m.module)

            [(key, value_dct)] = m.value.items()
            if not key.startswith('$'):
                raise Exception(f'Bad key: {key}')
            if key.startswith('$.'):
                key = f'${pkg_name}{key[1:]}'
                m = dc.replace(m, value={key: value_dct})

            lst.append(m)

        return lst

    #

    def _read_package_file_text(self, pkg_name: str, file_name: str) -> ta.Optional[str]:
        # importlib.resources.files actually imports the package - to avoid this, if possible, the file is read straight
        # off the filesystem.
        spec = importlib.util.find_spec(pkg_name)
        if (
                spec is not None and
                isinstance(spec.loader, importlib.machinery.SourceFileLoader) and
                spec.origin is not None and
                len(spec.submodule_search_locations or []) == 1 and
                os.path.basename(spec.origin) == '__init__.py' and
                os.path.isfile(spec.origin)
        ):
            file_path = os.path.join(os.path.dirname(spec.origin), file_name)
            if os.path.isfile(file_path):
                with open(file_path, encoding='utf-8') as f:
                    return f.read()

        t = importlib.resources.files(pkg_name).joinpath(file_name)
        if not t.is_file():
            return None
        return t.read_text('utf-8')

    MANIFESTS_FILE_NAME: ta.ClassVar[str] = '.manifests.json'

    def _load_package_uncached(self, pkg_name: str) -> ta.Optional[LoadedPackage]:
        src = self._read_package_file_text(pkg_name, self.MANIFESTS_FILE_NAME)
        if src is None:
            return None

        obj = json.loads(src)
        if not isinstance(obj, (list, tuple)):
            raise TypeError(obj)

        raw_lst = self._deserialize_raw_manifests(obj, pkg_name)

        ld_pkg = ManifestLoader.LoadedPackage(self, pkg_name)

        ld_man_lst: ta.List[ManifestLoader.LoadedManifest] = []
        for raw in raw_lst:
            ld_man = ManifestLoader.LoadedManifest(ld_pkg, raw)

            ld_man_lst.append(ld_man)

        ld_pkg._manifests = ld_man_lst  # noqa

        return ld_pkg

    def _load_package_locked(self, pkg_name: str) -> ta.Optional[LoadedPackage]:
        try:
            return self._loaded_packages[pkg_name]
        except KeyError:
            pass

        pkg = self._load_package_uncached(pkg_name)
        self._loaded_packages[pkg_name] = pkg
        return pkg

    def load_package(self, pkg_name: str) -> ta.Optional[LoadedPackage]:
        with self._lock:
            return self._load_package_locked(pkg_name)

    #

    def _instantiate_value(self, cls: type, **kwargs: ta.Any) -> ta.Any:
        if self._value_instantiator is not None:
            return self._value_instantiator(cls, **kwargs)
        else:
            return cls(**kwargs)

    def _instantiate_loaded_manifest(self, ld_man: LoadedManifest) -> ta.Any:
        [(cls_key, value_dct)] = ld_man.manifest.value.items()
        cls = self._load_class(cls_key)
        value = self._instantiate_value(cls, **value_dct)
        return value

    #

    # FIXME:
    # class LOAD_ALL:  # noqa
    #     def __new__(cls, *args, **kwargs):  # noqa
    #         raise TypeError
    #
    #     def __init_subclass__(cls, **kwargs):  # noqa
    #         raise TypeError

    def _load(
            self,
            *pkg_names: str,
            only: ta.Optional[ta.Iterable[type]] = None,
    ) -> ta.Sequence[LoadedManifest]:
        only_keys: ta.Optional[ta.Set]
        if only is not None:
            only_keys = set()
            for cls in only:
                if not (isinstance(cls, type) and dc.is_dataclass(cls)):
                    raise TypeError(cls)
                mod_name = cls.__module__
                mod_name = self._module_reverse_remap.get(mod_name, mod_name)
                only_keys.add(f'${mod_name}.{cls.__qualname__}')
        else:
            only_keys = None

        lst: ta.List[ManifestLoader.LoadedManifest] = []
        for pn in pkg_names:
            lp = self.load_package(pn)
            if lp is None:
                continue

            for m in lp.manifests:
                if only_keys is not None and m.class_key not in only_keys:
                    continue

                lst.append(m)

        return lst

    def load(
            self,
            *pkg_names: str,
            only: ta.Optional[ta.Iterable[type]] = None,
    ) -> ta.Sequence[LoadedManifest]:
        with self._lock:
            return self._load(
                *pkg_names,
                only=only,
            )

    #

    ENTRY_POINT_GROUP: ta.ClassVar[str] = 'omlish.manifests'

    _discovered_packages: ta.ClassVar[ta.Optional[ta.Sequence[str]]] = None

    @classmethod
    def discover_packages(cls) -> ta.Sequence[str]:
        if (x := cls._discovered_packages) is not None:
            return x

        # This is a fat dep so do it late.
        from importlib import metadata as importlib_metadata  # noqa

        x = [
            ep.value
            for ep in importlib_metadata.entry_points(group=cls.ENTRY_POINT_GROUP)
        ]

        cls._discovered_packages = x
        return x

    #

    def _scan_package_root_dir_uncached(
            self,
            root_dir: str,
    ) -> ta.Sequence[str]:
        pkgs: ta.List[str] = []

        for n in os.listdir(root_dir):
            if (
                    os.path.isdir(p := os.path.join(root_dir, n)) and
                    os.path.exists(os.path.join(p, '__init__.py'))
            ):
                pkgs.append(n)

        return pkgs

    def _scan_package_root_dir_locked(
            self,
            root_dir: str,
    ) -> ta.Sequence[str]:
        try:
            return self._scanned_package_root_dirs[root_dir]
        except KeyError:
            pass

        ret = self._scan_package_root_dir_uncached(root_dir)
        self._scanned_package_root_dirs[root_dir] = ret
        return ret

    def _scan_package_root_dir(
            self,
            root_dir: str,
    ) -> ta.Sequence[str]:
        with self._lock:
            return self._scan_package_root_dir_locked(root_dir)

    def scan_or_discover_packages(
            self,
            *,
            specified_root_dirs: ta.Optional[ta.Sequence[str]] = None,
            fallback_root_dir: ta.Optional[str] = None,
    ) -> ta.Sequence[str]:
        pkgs: list[str] = []

        if specified_root_dirs is not None:
            if isinstance(specified_root_dirs, str):
                raise TypeError(specified_root_dirs)

            for r in specified_root_dirs:
                pkgs.extend(self._scan_package_root_dir(r))

        else:
            pkgs.extend(self.discover_packages())

            if not pkgs and fallback_root_dir is not None:
                pkgs.extend(self._scan_package_root_dir(fallback_root_dir))

        return pkgs



# ruff: noqa: UP045
# @omlish-lite
import threading
import typing as ta

from ..lite.marshal import unmarshal_obj
from .loading import ManifestLoader


##


class GlobalManifestLoader:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    def __init_subclass__(cls, **kwargs):  # noqa
        raise TypeError

    ##

    _lock: ta.ClassVar[threading.RLock] = threading.RLock()

    _instance: ta.ClassVar[ta.Optional[ManifestLoader]] = None

    @classmethod
    def instance(cls) -> ManifestLoader:
        if (inst := cls._instance) is None:
            with cls._lock:
                if (inst := cls._instance) is None:
                    inst = cls._instance = ManifestLoader(**cls.default_kwargs())

        return inst

    @classmethod
    def initialize(cls, **kwargs: ta.Any) -> ManifestLoader:
        with cls._lock:
            if cls._instance is not None:
                raise Exception(f'{cls.__name__} already initialized')

            inst = cls._instance = ManifestLoader(**kwargs)

        return inst

    ##

    @classmethod
    def default_value_instantiator(cls, obj_cls: type, **kwargs: ta.Any) -> ta.Any:
        return unmarshal_obj(kwargs, obj_cls)

    @classmethod
    def default_kwargs(cls) -> ta.Mapping[str, ta.Any]:
        return dict(
            value_instantiator=cls.default_value_instantiator,
        )

    ##

    @classmethod
    def load(
            cls,
            *pkg_names: str,
            only: ta.Optional[ta.Iterable[type]] = None,
    ) -> ta.Sequence[ManifestLoader.LoadedManifest]:
        return cls.instance().load(
            *pkg_names,
            only=only,
        )
